﻿Imports System.Runtime.InteropServices
Imports System.Security
Imports System.Drawing.Drawing2D

Public Class Form1
    <StructLayout(LayoutKind.Sequential)> _
    Private Structure NativeMessage
        Public handle As IntPtr
        Public msg As UInteger
        Public wParam As IntPtr
        Public lParam As IntPtr
        Public time As UInteger
        Public p As System.Drawing.Point
    End Structure
    Private Declare Auto Function PeekMessage Lib "user32.dll" ( _
        ByRef lpMsg As NativeMessage, _
        ByVal hWnd As IntPtr, _
        ByVal wMsgFilterMin As UInteger, _
        ByVal wMsgFilterMax As UInteger, _
        ByVal flags As UInteger _
    ) As Boolean
    Private Const WM_MOUSEFIRST As UInteger = &H200
    Private Const WM_MOUSELAST As UInteger = &H20D
    Private Const PM_REMOVE As Integer = &H1

    ' Flush all pending mouse events.
    Private Sub FlushMouseMessages()
        Dim msg As NativeMessage
        ' Repeat until PeekMessage returns false.
        While (PeekMessage(msg, IntPtr.Zero, WM_MOUSEFIRST, WM_MOUSELAST, PM_REMOVE))

        End While
    End Sub

    Private Const Delay As Integer = 250

    ' Geometry.
    Private Const NumDisks As Integer = 5
    Private Const Y1 As Integer = 10
    Private Const Y2 As Integer = 50
    Private Const DiskThickness As Integer = 20
    Private X1, X2, X3, Y3 As Integer

    Private PegLocation(3) As Rectangle

    ' Peg stacks. Each disk is represented by its width.
    Private PegStack(3) As Stack(Of Integer)

    ' The peg that currently holds the disks.
    Private OccupiedPegNum As Integer = 0

    ' True if we are performing a move.
    Private Moving As Boolean = False

    ' The position of the disk that is moving.
    Private MovingDiskRect As Rectangle

    ' Build the controls.
    Private Sub Form1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Me.DoubleBuffered = True

        ' Make the peg stacks.
        For i As Integer = 0 To 2
            PegStack(i) = New Stack(Of Integer)()
        Next i

        ' Calculate the widest disk's width.
        Dim disk_width As Integer = (NumDisks + 2) * DiskThickness

        ' See how big the form must be.
        ClientSize = New Size(3 * disk_width + 4 * DiskThickness, ClientSize.Height)

        ' Figure out where the pegs must be.
        X1 = CInt(DiskThickness + disk_width / 2)
        X2 = X1 + disk_width + DiskThickness
        X3 = X2 + disk_width + DiskThickness
        Y3 = ClientSize.Height - 10

        PegLocation(0) = New Rectangle(CInt(X1 - DiskThickness / 2), Y2, DiskThickness, Y3 - Y2)
        PegLocation(1) = New Rectangle(CInt(X2 - DiskThickness / 2), Y2, DiskThickness, Y3 - Y2)
        PegLocation(2) = New Rectangle(CInt(X3 - DiskThickness / 2), Y2, DiskThickness, Y3 - Y2)

        ' Make the disks.
        For i As Integer = 0 To NumDisks - 1
            PegStack(0).Push(disk_width)
            disk_width -= DiskThickness
        Next i
        OccupiedPegNum = 0
    End Sub

#Region "Mouse"

    ' If the mouse is over a clickable peg, display a crosshair cursor.
    Private Sub Form1_MouseMove(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles MyBase.MouseMove
        Dim peg_under_mouse As Integer = PegAtLocation(e.Location)
        If (peg_under_mouse < 0) Then
            Me.Cursor = Cursors.Default
        Else
            Me.Cursor = Cursors.Cross
        End If
    End Sub

    ' Return the index of the peg at this location.
    ' Return -1 if the location is not on any peg
    ' or if the peg is not clickable.
    Private Function PegAtLocation(ByVal location As Point) As Integer
        If (Moving) Then Return -1 ' Not clickable if we're moving.

        For i As Integer = 0 To 2
            If (PegLocation(i).Contains(location)) Then
                ' Not clickable if this is the occupied peg.
                If (i = OccupiedPegNum) Then Return -1
                Return i
            End If
        Next i
        Return -1
    End Function

    ' Move disks to this peg.
    Private Sub Form1_MouseClick(ByVal sender As System.Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles MyBase.MouseClick
        If (Moving) Then Return
        Dim peg_under_mouse As Integer = PegAtLocation(e.Location)
        If (peg_under_mouse < 0) Then Return

        Moving = True
        Me.Cursor = Cursors.WaitCursor
        Application.DoEvents()

        Dim other_peg As Integer = 0
        If (OccupiedPegNum <> 1 AndAlso peg_under_mouse <> 1) Then other_peg = 1
        If (OccupiedPegNum <> 2 AndAlso peg_under_mouse <> 2) Then other_peg = 2

        MoveDisks(OccupiedPegNum, peg_under_mouse, other_peg, NumDisks)
        OccupiedPegNum = peg_under_mouse

        Me.Cursor = Cursors.Default
        Moving = False
        Me.Refresh()
        FlushMouseMessages()   ' Discard any pending clicks.
    End Sub

#End Region ' Mouse

    ' Move num_disks disks from peg from_peg to peg to_peg.
    Private Sub MoveDisks(ByVal from_peg As Integer, ByVal to_peg As Integer, ByVal other_peg As Integer, ByVal num_disks As Integer)
        ' See if we have more than one disk to move.
        If (num_disks > 1) Then
            ' We have more than one disk to move.
            ' Recursively move all but one disk to the other peg.
            MoveDisks(from_peg, other_peg, to_peg, num_disks - 1)
        End If

        ' Move the remaining disk.
        AnimateDiskMove(from_peg, to_peg)

        If (num_disks > 1) Then
            ' Recursively move the other disks back where they belong.
            MoveDisks(other_peg, to_peg, from_peg, num_disks - 1)
        End If
    End Sub

#Region "Drawing"

    ' Draw the pegs and disks.
    Private Sub Form1_Paint(ByVal sender As System.Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles MyBase.Paint
        e.Graphics.Clear(Me.BackColor)
        e.Graphics.SmoothingMode = SmoothingMode.AntiAlias

        DrawPegs(e.Graphics)
        DrawDisks(e.Graphics)
    End Sub

    ' Draw the pegs.
    Private Sub DrawPegs(ByVal gr As Graphics)
        For i As Integer = 0 To 2
            gr.FillRectangle(Brushes.Orange, PegLocation(i))
            gr.DrawRectangle(Pens.Red, PegLocation(i))
        Next i
    End Sub

    ' Draw the disks.
    Private Sub DrawDisks(ByVal gr As Graphics)
        ' For each peg...
        For peg As Integer = 0 To 2
            ' For each disk on the peg...
            Dim y As Integer = Y3 - 2 - DiskThickness
            For Each disk_width As Integer In PegStack(peg).ToArray().Reverse()
                ' Draw the disk.
                Dim disk_rect As New Rectangle( _
                    CInt(PegLocation(peg).X + DiskThickness / 2 - disk_width / 2), _
                    y, disk_width, DiskThickness)
                gr.FillRectangle(Brushes.LightGreen, disk_rect)
                gr.DrawRectangle(Pens.Green, disk_rect)
                y -= DiskThickness + 2
            Next disk_width
        Next peg

        ' If we are currently moving a disk, draw it.
        If (Moving) Then
            gr.FillRectangle(Brushes.LightBlue, MovingDiskRect)
            gr.DrawRectangle(Pens.Blue, MovingDiskRect)
        End If
    End Sub

    ' Animate moving a disk from peg from_peg to peg to_peg.
    Private Sub AnimateDiskMove(ByVal from_peg As Integer, ByVal to_peg As Integer)
        ' Remove the disk so Paint doesn't draw it.
        Dim disk_width As Integer = PegStack(from_peg).Pop()

        ' Move the disk above from_peg.
        Dim y As Integer = PegLocation(from_peg).Bottom - _
            (PegStack(from_peg).Count + 1) * (DiskThickness + 2)
        Dim x As Integer = CInt(PegLocation(from_peg).X + DiskThickness / 2 - disk_width / 2)
        MovingDiskRect = New Rectangle(x, y, disk_width, DiskThickness)

        ' Move the disk up above the peg.
        Dim to_y As Integer = 10
        AnimateMovement(x, to_y)

        ' Move the disk over the new peg.
        Dim to_x As Integer = CInt(PegLocation(to_peg).X + DiskThickness / 2 - disk_width / 2)
        AnimateMovement(to_x, to_y)

        ' Move the disk down onto the peg.
        to_y = PegLocation(to_peg).Bottom - _
            (PegStack(to_peg).Count + 1) * (DiskThickness + 2)
        AnimateMovement(to_x, to_y)

        ' Redraw.
        Me.Refresh()

        ' Add the disk to its new peg.
        MovingDiskRect.X = -1
        PegStack(to_peg).Push(disk_width)
    End Sub

    ' Move the moving disk to this location.
    Private Sub AnimateMovement(ByVal end_x As Integer, ByVal end_y As Integer)
        Dim start_x As Integer = MovingDiskRect.X
        Dim start_y As Integer = MovingDiskRect.Y

        Const pixels_per_second As Integer = 400
        Dim dx As Single = end_x - MovingDiskRect.X
        Dim dy As Single = end_y - MovingDiskRect.Y
        Dim dist As Single = CSng(Math.Sqrt(dx * dx + dy * dy))

        ' Calculate distance moved per second.
        dx = pixels_per_second * dx / dist
        dy = pixels_per_second * dy / dist

        ' See how long the total move will take.
        Dim seconds As Single = dist / pixels_per_second
        Dim start_time As Date = Date.Now

        ' Start moving.
        Do
            ' Redraw.
            Me.Refresh()

            ' Wait a little while.
            System.Threading.Thread.Sleep(10)

            ' See how much time has passed.
            Dim elapsed As TimeSpan = DateTime.Now - start_time
            If (elapsed.TotalSeconds > seconds) Then Exit Do

            ' Update the rectangle's position.
            MovingDiskRect.X = CInt(start_x + elapsed.TotalSeconds * dx)
            MovingDiskRect.Y = CInt(start_y + elapsed.TotalSeconds * dy)
        Loop

        MovingDiskRect.X = end_x
        MovingDiskRect.Y = end_y
    End Sub

#End Region ' Drawing

End Class
